@@ -37,6 +37,8 @@ ERROR_ID_MESSAGE = (
     "and must be between 1 and 36 characters."
 )
 
+ERROR_EMPTY_PAYLOAD = "Invalid payload given, it should contains a least one agent"
+
 
 def current_time_ms():
     return int(round(time.time() * 1000))
@@ -93,6 +95,10 @@ class Client(object):
             cfg.get("decisionTreeRetrievalTimeout"), int
         ):
             cfg["decisionTreeRetrievalTimeout"] = 1000 * 60 * 5  # 5 minutes
+        if cfg.get("boostingDecisionRetrievalTimeout") is not False and not isinstance(
+            cfg.get("boostingDecisionRetrievalTimeout"), int
+        ):
+            cfg["boostingDecisionRetrievalTimeout"] = 1000 * 60 * 5  # 5 minutes
         if not isinstance(cfg.get("url"), str):
             cfg["url"] = "https://beta.craft.ai"
         if cfg.get("url").endswith("/"):
@@ -665,6 +671,13 @@ class Client(object):
         # Raises an error when agent_id is invalid
         self._check_entity_id(agent_id)
 
+        # Suggest pandas client if a dataframe is provided
+        if hasattr(operations, "shape"):
+            CraftAiBadRequestError(
+                """A dataframe of operations has been provided,
+                the pandas Client handle such type of data"""
+            )
+
         # Extra header in addition to the main session's
         ct_header = {"Content-Type": "application/json; charset=utf-8"}
         offset = 0
@@ -701,7 +714,7 @@ class Client(object):
             "added_operations_count": added_operations_count,
         }
 
-    def _add_agents_operations_bulk(self, chunked_data):
+    def _add_agents_operations_bulk(self, chunked_data, invalid_agents):
         """Tool for the function add_agents_operations_bulk. It send the requests to
         add the operations to the agents.
 
@@ -755,6 +768,10 @@ class Client(object):
         if responses == []:
             raise CraftAiBadRequestError("Invalid or empty set of operations given")
 
+        if invalid_agents != []:
+            for invalid_agent in invalid_agents:
+                responses.append(invalid_agent)
+
         return responses
 
     def add_agents_operations_bulk(self, payload):
@@ -776,7 +793,7 @@ class Client(object):
         referenced non existing agents or one of the operations is invalid.
         """
         # Check all ids, raise an error if all ids are invalid
-        valid_indices, _, _ = self._check_entity_id_bulk(payload)
+        valid_indices, _, invalid_agents = self._check_entity_id_bulk(payload)
         valid_payload = [payload[i] for i in valid_indices]
 
         chunked_data = []
@@ -802,7 +819,7 @@ class Client(object):
         if current_chunk:
             chunked_data.append(current_chunk)
 
-        return self._add_agents_operations_bulk(chunked_data)
+        return self._add_agents_operations_bulk(chunked_data, invalid_agents)
 
     def _get_agent_operations_pages(self, url, ops_list):
         if url is None:
@@ -1033,8 +1050,159 @@ class Client(object):
 
     @staticmethod
     def decide(tree, *args):
+        for argument in args:
+            # Suggest pandas client if a dataframe is provided
+            if hasattr(argument, "shape"):
+                raise CraftAiBadRequestError(
+                    """A dataframe of operations has been provided,
+                    the pandas Client handle such type of data"""
+                )
         return Interpreter.decide(tree, args)
 
+    ####################
+    # Boosting methods #
+    ####################
+
+    def get_agent_boosting_decision(self, agent_id, from_ts, to_ts, context):
+        """Get boosting decision.
+
+        :param str agent_id: the id of the agent whose tree to get. It
+        must be an str containing only characters in "a-zA-Z0-9_-" and
+        must be between 1 and 36 characters.
+        :param int from_ts: The boosting model will be built from this
+        timestamp.
+        :param int to_ts: The boosting model will be built until this
+        timestamp.
+        :param dictionary context: Contains dictionnaries that has the
+        form given in the craft_ai documentation and the configuration
+        of the agent.
+
+        :return: boosting_decision.
+        :rtype: dict.
+
+        :raises CraftAiLongRequestTimeOutError: if the API doesn't get
+        the tree in the time given by the configuration.
+        """
+
+        entity_type = "agents"
+
+        # Convert datetime to timestamp
+        if isinstance(from_ts, datetime.datetime):
+            from_ts = time.mktime(from_ts.timetuple())
+
+        if isinstance(to_ts, datetime.datetime):
+            to_ts = time.mktime(to_ts.timetuple())
+
+        window = [from_ts, to_ts]
+        # Raises an error when agent_id is invalid
+        self._check_entity_id(agent_id)
+        if self._config["boostingDecisionRetrievalTimeout"] is False:
+            # Don't retry
+            return self._get_entity_boosting_decision(
+                agent_id, window, entity_type, context
+            )
+
+        start = current_time_ms()
+        while True:
+            now = current_time_ms()
+            if now - start > self._config["boostingDecisionRetrievalTimeout"]:
+                # Client side timeout
+                raise CraftAiLongRequestTimeOutError()
+            try:
+                return self._get_entity_boosting_decision(
+                    agent_id, window, entity_type, context
+                )
+            except CraftAiLongRequestTimeOutError:
+                # Do nothing and continue.
+                continue
+
+    def get_generator_boosting_decision(self, generator_id, from_ts, to_ts, context):
+        """Get boosting decision.
+
+        :param str generator_id: the id of the agent whose tree to get. It
+        must be an str containing only characters in "a-zA-Z0-9_-" and
+        must be between 1 and 36 characters.
+        :param int from_ts: The boosting model will be built from this
+        timestamp.
+        :param int to_ts: The boosting model will be built until this
+        timestamp.
+        :param dictionary context: Contains dictionnaries that has the
+        form given in the craft_ai documentation and the configuration
+        of the agent.
+
+        :return: boosting_decision.
+        :rtype: dict.
+
+        :raises CraftAiLongRequestTimeOutError: if the API doesn't get
+        the tree in the time given by the configuration.
+        """
+
+        entity_type = "generators"
+
+        # Convert datetime to timestamp
+        if isinstance(from_ts, datetime.datetime):
+            from_ts = time.mktime(from_ts.timetuple())
+
+        if isinstance(to_ts, datetime.datetime):
+            to_ts = time.mktime(to_ts.timetuple())
+
+        window = [from_ts, to_ts]
+        # Raises an error when generator_id is invalid
+        self._check_entity_id(generator_id)
+        if self._config["boostingDecisionRetrievalTimeout"] is False:
+            # Don't retry
+            return self._get_entity_boosting_decision(
+                generator_id, window, entity_type, context
+            )
+
+        start = current_time_ms()
+        while True:
+            now = current_time_ms()
+            if now - start > self._config["boostingDecisionRetrievalTimeout"]:
+                # Client side timeout
+                raise CraftAiLongRequestTimeOutError()
+            try:
+                return self._get_entity_boosting_decision(
+                    generator_id, window, entity_type, context
+                )
+            except CraftAiLongRequestTimeOutError:
+                # Do nothing and continue.
+                continue
+
+    def _get_entity_boosting_decision(self, entity_id, window, entity_type, context):
+        """Tool for the function get_agent_boosting_decision
+        and get_generator_boosting_decision.
+
+        :param str entity_id: the id of the agent whose tree to get. It
+        must be an str containing only characters in "a-zA-Z0-9_-" and
+        must be between 1 and 36 characters.
+        :param array window: Time window over which samples are selected
+        to build the model
+        :param entity_type: the entity type corresponding to the entity_uri
+        agents or generators.
+        :param: Contains dictionnaries that has the
+        form given in the craft_ai documentation and the configuration
+        of the entity.
+
+        :return: decision.
+        :rtype: dict.
+        """
+
+        ct_header = {"Content-Type": "application/json; charset=utf-8"}
+        # If we give no timestamp the default behaviour is to give
+        # the tree from the latest timestamp
+        req_url = "{}/{}/{}/boosting/decision".format(
+            self._base_url, entity_type, entity_id
+        )
+        payload = {"timeWindow": window, "context": context}
+        json_pl = json.dumps(payload)
+
+        resp = self._requests_session.post(req_url, headers=ct_header, data=json_pl)
+
+        decision = self._decode_response(resp)
+
+        return decision
+
     @staticmethod
     def _parse_body(response):
         try:
@@ -1187,6 +1355,9 @@ class Client(object):
                 else:
                     valid_entity_indices.append(index)
 
+        if len(payload) == 0:
+            raise CraftAiBadRequestError(ERROR_EMPTY_PAYLOAD)
+
         if len(invalid_entity_indices) == len(payload):
             raise CraftAiBadRequestError(ERROR_ID_MESSAGE)
 
